#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

*******************************************************************************
calc.py  :  url requests to stoneage.hzdr.de and calculation of erosion rate 

    Copyright (C) 2023  Konstanze Stübner, kstueb@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*******************************************************************************

General functions:
- get_textline() : Cast 'sample' and 'topo' data to textstring for the online
    calculator. Input dict ot pd.Series. If possible, missing values are filled
    with default values from params.Params.
    (get_textblock() for point-based ages & erosion rates is part of utils.py)
- get_ages_from_server(), get_erates_from_server(), get_NofE_from_server() :
    Get pd.DataFrame, diagnosis and version for any string sent to the
    online calculator.

Specialized function:
- get_E() : dict w/ keys 'St', 'Lm', 'LSDn' for a single sample-nuclide pair.
- get_NofE_FullTable() : Full table of 'NpredXX' for 'topostats' and suite of
    'erates'. Results are weighted by 'wt' ('XX_wt').
- get_NofE() : Wrapper for _FullTable(): groups data by 'E_cmyr'.

The work horse:
- poly_E_results() : -> (E, delE, NofE, error).
    Returns error=['minE too high'] or ['maxE too low']
    or a list of one or several other errors.

"""

import numpy as np
import pandas as pd
from numbers import Number

from copy import deepcopy
from scipy import optimize

import urllib.parse
import urllib.request

from riversand.params import Params


def get_textline(sample:pd.Series, topo:pd.Series, shielding) -> str:
    """
    Cast sample and topo data to textstring for the online calculator. Input
    data are validated by utils.validate_*().

    Parameters
    ----------
    sample : pandas Series or dict
        > sample_data = S.data.iloc[n]
        Missing data are replaced by default values, faulty data raise ValueError.
    topo : pandas Series or dict 
        'lat', 'long', 'elevation' (and 'shielding' if shielding='topo').
        > topo_data = topotable.iloc[n]
        or:
        > summary['elevation'] = summary['elevLo']
        > topo_data = summary
    shielding : str or float
        'sample', 'topo' or a numeric shielding factor.
        Defines whether shielding is taken from sample or raster data.
    
    Returns
    -------
    textline : str
        String corresponding to one location and one set of sample data (Be-10
        or Al-26). Re-standardized nuclide concentrations.
        
    Raises
    ------
    TypeError
        Input data format is invalid.
    ValueError
        Data cannot be validated.
        
    """
    
    sample = deepcopy(sample)
    topo = deepcopy(topo)
    
    from riversand.utils import validate_topo, validate_nuclide, validate_sample
    #from riversand.utils import restandardize_item
        
    if isinstance(sample, pd.DataFrame):
        if len(sample)!=1:
            raise TypeError("get_textline() argument 'sample' must be length 1")
        sample = sample.iloc[0] # recast to pd.Series
            
        if isinstance(sample, pd.Series):
            sample = sample.to_dict()
        
        if not isinstance(sample, dict):
            raise TypeError("get_textline() argument 'sample' must be dict or pandas Series")
    
    if isinstance(topo, pd.DataFrame):
        if len(topo)!=1:
            raise TypeError("get_textline() argument 'topo' must be length 1")
        topo = topo.iloc[0] # recast to pd.Series
            
        if isinstance(topo, pd.Series):
            topo = topo.to_dict()
        
        if not isinstance(topo, dict):
            raise TypeError("get_textline() argument 'topo' must be dict or pandas Series")
    
    if shielding=='sample':
        try:
            sf = float(sample['shielding'])
        except:
            sf = np.nan
        finally:
            if np.isnan(sf):
                raise ValueError("Invalid shielding='sample': no shielding "+
                                 "defined in sample data")
    if shielding=='topo':
        try:
            sf = float(topo['shielding'])
        except:
            sf = np.nan
        finally:
            if np.isnan(sf):
                raise ValueError("Invalid shielding='topo': no shielding "+
                                 "defined in topo data")
            
    if isinstance(shielding, Number):
        sample['shielding'] = shielding
        shielding = 'sample'
    
    if shielding=='sample':
        try:
            out1 = validate_topo(topo)
            out2 = validate_nuclide(sample)
            out3 = validate_sample(sample, shielding=True)
        except ValueError as e:
            raise e
            
    elif shielding=='topo':
        try:
            out1 = validate_topo(topo, shielding=True)
            out2 = validate_nuclide(sample)
            out3 = validate_sample(sample)
        except ValueError as e:
            raise e
    else:
        raise ValueError("Invalid shielding: must be 'topo', 'sample' or numeric")
    
    out = {**out1, **out2, **out3}
    
    #out = restandardize_item(out)
            
    textline = "{} {:.3e} {:.3e} {:.4e} {} {} {} {:.3e} {:.3e} {} ; {} {} {} {} {} {} ;".format(
        out['name'], out['lat'], out['long'], out['elevation'], out['press_flag'],
        out['thickness'], out['density'], out['shielding'], out['erate'], out['year'],
        out['name'], out['nuclide'], out['mineral'], out['N'], out['delN'], out['standardization'])
    return textline



def get_ages_from_server(
    textblock:str,
    url:str=None,
    norm=False
    ) -> (pd.DataFrame, str, dict):
    """
    Get ages in yr from online calculator. Example of usage:
        
    > df = pd.read_excel('./folder/file.xlsx')  # read from excel spreadsheet
    > df = import_data(df)                      # recast for online calculator / validate
    > textblock = get_textblock(df)
    > ages, diagnostics, version = get_ages_from_server(textblock)
    or
    > ages, diagnostics, version = get_ages_from_server(textblock, norm=True)
    
    Parameters
    ----------
    textblock : str
        Any valid string of a single sample or multiple samples.
    norm : bool
        If True the results include the normalized values used for plotting. 
    
    Returns
    -------
    ages : pd.DataFrane
        Age results; one row per sample and nuclide.
        Keys are: 'name', 'nuclide',
            'St_age', 'St_interr', 'St_exterr',
            'Lm_age', 'Lm_interr', 'Lm_exterr',
            'LSDn_age', 'LSDn_interr', 'LSDn_exterr',
    diagnostics : str
        "No response from {}".format(url)
        "No diagnostics" as returned by the server
        "Sample PH-1 -- N10quartz appears to be saturated WRT Stone(2000) SF at this erosion rate."  ...or similar
        "validate_v2_input.m: Wrong total number of data elements"  ...or similar if the server does not accept the input string
        "Empty string returned from the server. This may happen if at least one sample has an extremely low nuclide concentration."
    version :  dict
        Returned by the server.
       
    Raises
    ------
    TypeError
        "'textblock' is empty string"
        
    """
    
    # rename from xml-string to online output
    out_cols = {'sample': 'name',
                'nuclide': 'nuclide',
                't_St': 'St_age',
                'delt_int_St': 'St_interr',
                'delt_ext_St': 'St_exterr',
                't_Lm': 'Lm_age',
                'delt_int_Lm': 'Lm_interr',
                'delt_ext_Lm': 'Lm_exterr',
                't_LSDn': 'LSDn_age',
                'delt_int_LSDn': 'LSDn_interr',
                'delt_ext_LSDn': 'LSDn_exterr',
                #'Ag_age', 'Ag_interr', 'Ag_exterr'
                }
           
       
    # additional columns for norm=True:
    norm_cols = ['Nnorm_St', 'delNnorm_int_St', 'delNnorm_ext_St',
                 'Nnorm_Lm', 'delNnorm_int_Lm', 'delNnorm_ext_Lm',
                 'Nnorm_LSDn', 'delNnorm_int_LSDn', 'delNnorm_ext_LSDn']
    version = {}
    
    ages = pd.DataFrame(columns=out_cols.values()) # empty DataFrame if no response from server
    if norm: ages[norm_cols] = None # append norm columns at end
    
    from riversand import xml
    from riversand.utils import get_textblock
    
    if url is None: url = Params.url

    if isinstance(textblock, pd.DataFrame):
        try:
            textblock = get_textblock(textblock)
        except ValueError:
            raise TypeError("cannot convert input to textblock for online calculator. "+
                            "Try get_textblock() for details.") from None
            # This should never happen; get_textblock() uses import_data2() to
            # validate input and returns an empty string if data is faulty

    if len(textblock.replace(' ',''))==0:
        raise TypeError("'textblock' is empty string")
    
    # request ages from server
    request_vars = {"mlmfile" : "age_input_v3", #erosion_input_v3
                    "reportType" : "XML",
                    "resultType" : "long",
                    "summary" : "no",
                    "plotFlag" : "no",
                    "text_block" : textblock }

    form_data = urllib.parse.urlencode(request_vars).encode('ascii')
    try:
        result = urllib.request.urlopen(url, form_data)
    except urllib.error.URLError:
        diagnostics = "No response from {}".format(url) ## is this actually a valid return string?
    else:    
        result_xml = result.read()
        ages, diagnostics, version = xml.ages_from_xml(result_xml, norm)
    
    ages = ages.rename(columns=out_cols)
    ages = ages[[c for c in ages if c not in norm_cols] + # keep age columns at beginning
                [c for c in norm_cols if c in ages]] # move norm columns to end
    
    return ages, diagnostics, version



def get_erates_from_server(
        textblock:str,
        url:str=None
        ) -> (pd.DataFrame, str, dict):
    """
    Get erosion rates in g/cm2/yr from online calculator. Example of usage:
    
    > df = pd.read_excel('./folder/file.xlsx')  # read from excel spreadsheet
    > df = import_data(df)                      # recast for online calculator / validate
    > textblock = get_textblock(df)
    > erates, diagnostics, version = get_erates_from_server(textblock)
    
    Parameters
    ----------
    textblock : str
        Any valid string of a single sample or multiple samples.
        
    Returns
    -------
    erates : pd.DataFrame
        Erosion rate results; one row per sample and nuclide.
        Keys are: 'name', 'nuclide', 'density',
                  'E_St', 'delE_int_St', 'delE_ext_St',
                  'E_Lm', 'delE_int_Lm', 'delE_ext_Lm',
                  'E_LSDn', 'delE_int_LSDn', 'delE_ext_LSDn' 
    diagnostics : str
        "No response from {}".format(url)
        "No diagnostics" as returned by the server
        "Sample PH-1 -- N10quartz appears to be saturated WRT Stone(2000) SF at this erosion rate."  ...or similar
        "validate_v2_input.m: Wrong total number of data elements"  ...or similar if the server does not accept the input string
        "Empty string returned from the server. This may happen if at least one sample has an extremely low nuclide concentration."
    version : dict
        Returned by the server.
        
    Raises
    ------
    TypeError
        "'textblock' is empty string"
        
    """
    
    # rename from xml-string to online output
    out_cols = {'sample': 'name',
                'nuclide': 'nuclide',
                'density': 'density',
                'E_gcm2_St': 'E_St',
                'delE_int_gcm2_St': 'delE_int_St',
                'delE_ext_gcm2_St': 'delE_ext_St',
                'E_gcm2_Lm': 'E_Lm',
                'delE_int_gcm2_Lm': 'delE_int_Lm',
                'delE_ext_gcm2_Lm': 'delE_ext_Lm',
                'E_gcm2_LSDn': 'E_LSDn',
                'delE_int_gcm2_LSDn': 'delE_int_LSDn',
                'delE_ext_gcm2_LSDn': 'delE_ext_LSDn'}
    version = {}
    
    erates = pd.DataFrame(columns=out_cols.values()) # empty DataFrame if no response from server
    
    from riversand import xml
    from riversand.utils import get_textblock

    if url is None: url = Params.url
        
    if isinstance(textblock, pd.DataFrame):
        try:
            textblock = get_textblock(textblock)
        except ValueError:
            raise TypeError("cannot convert input to textblock for online calculator. "+
                            "Try get_textblock() for details.") from None
            # This should never happen; get_textblock() uses import_data2() to
            # validate input and returns an empty string if data is faulty
            
    if len(textblock.replace(' ',''))==0:
        raise TypeError("'textblock' is empty string")
    
    
    # request erosion rate from server
    request_vars = {"mlmfile" : "erosion_input_v3",
                    "reportType" : "XML",
                    "resultType" : "long",
                    "summary" : "no",
                    "plotFlag" : "no",
                    "text_block" : textblock }
    
    form_data = urllib.parse.urlencode(request_vars).encode('ascii')
    try:        
        result = urllib.request.urlopen(url, form_data)
    except urllib.error.URLError:
        diagnostics = "No response from {}".format(url)
    else:    
        result_xml = result.read()
        erates, diagnostics, version = xml.erates_from_xml(result_xml)
    
    erates = erates.rename(columns=out_cols)
    
    return erates, diagnostics, version
    

def get_E(textline:str, url:str=None) -> dict:
    """
    Get erosion rates in cm/yr from online calculator for
    a SINGLE sample and nuclide.
    

    Returns
    -------
    ret_dict : dict
        keys 'St', 'Lm', 'LSDn'; erosion rate in cm/yr
    
    Raises RuntimeError if diagnostics is anything but "No diagnostics". 
       
    """

    if url is None: url = Params.url
    
    dfE = pd.DataFrame()
    ret_dict = {'St' : None, 'Lm' : None, 'LSDn' : None}
        
    # erates in g/cm2/yr
    dfE, diagnostics, version = get_erates_from_server(
        textline, url=url)
    
    if "saturated" in diagnostics:
        raise RuntimeError("get_E() : sample appears to be saturated")
    elif "Empty string returned from the server" in diagnostics:
        raise RuntimeError("get_E() : nuclide concentration too low for calculation")
    elif "No response from" in diagnostics:
        raise RuntimeError("get_E() : {}".format(diagnostics))
        
        
    if len(dfE)!=1: 
        raise RuntimeError("get_E() : invalid results, probably from invalid "+
                           "argument 'textline'; must be a single sample and nuclide")
        
    try:
        ret_dict = {'St' : float(dfE.loc[0,'E_St']) / float(dfE.loc[0,'density']),
                    'Lm' : float(dfE.loc[0,'E_Lm']) / float(dfE.loc[0,'density']),
                    'LSDn' : float(dfE.loc[0,'E_LSDn']) / float(dfE.loc[0,'density']),
                    #'diagnostics' : diagnostics
                    }
    except:
        diagnostics = "Cannot compute E [g/cm2] from server output"
    
    if diagnostics!="No diagnostics":
        raise RuntimeError("get_E() : {}".format(diagnostics))
    return ret_dict

    
def get_NofE_from_server(
        textline:str,
        url:str=None
        ) -> (pd.DataFrame, str, dict):
    """
    Get predicted nuclide concentrations in at/g from online calculator.
    Note that result is independent of nuclide concentration and standardization.
    
    Parameters
    ----------
    textline : str
        Any valid textblock of multiple samples / nuclide / erates.
        
    Returns
    -------
    dfN : pd.DataFrame
        Nuclide concentration results; one row per sample and nuclide.
        Keys are: 'name', 'nuclide', 'E_cmyr',
                  'NpredSt', 'NpredLm', 'NpredLSDn' 
    diagnostics : str
        "No response from {}".format(url)
        "No diagnostics" as returned by the server
    version : dict
        Returned by the server.
        
    """
    
    dfN = pd.DataFrame(columns = ['name', 'nuclide', 'E_cmyr',
                                  'NpredSt', 'NpredLm', 'NpredLSDn'])
    version = {}
    
    from riversand import xml
    
    if url is None: url = Params.url
        
    if not isinstance(textline, str):
        raise TypeError("get_NofE_from_server() 'textline' must be string")
    if len(textline)==0:
        raise TypeError("'textline' is empty string")

    # request erosion rate from server
    request_vars = {"mlmfile" : "NofE_input_v3", "text_block" : textline}
    form_data = urllib.parse.urlencode(request_vars).encode('ascii')
    
    try:
        result = urllib.request.urlopen(url, form_data)
    except urllib.error.URLError:
        diagnostics = "No response from {}".format(url)
    else:    
        result_xml = result.read() # extract result
        dfN, diagnostics, version = xml.NofE_from_xml(result_xml)
    
    return dfN, diagnostics, version
    
    
    
def get_NofE_FullTable(
        sample_data:pd.Series,
        topostats: pd.DataFrame,
        shielding:str,
        erates:np.ndarray,
        url:str=None
    ) -> (pd.DataFrame, str, dict):
    """
    Get predicted nuclide concentrations for given topographic statistics
    for a single sample and a suite of erates.
    - lat, long, elevation from topostats
    - N, delN, press_flag, density, etc. from sample_data
    Sample names are auto-generated (one for each topostats entry).
    Sample thickness is 0.
    Note that result is independent of nuclide concentration and standardization
    specified in 'sample_data'.

    Parameters
    ----------
    sample_data : pd.Series or dict
        Single sample.
    topostats : pd.DataFrame
        Elevation-binned topo data.
    shielding : str or float
        'topo', 'sample' or numeric.
    erates : iterable or float.
        (Suite of) erosion rates in cm/yr.
    
    Returns
    ----------
    dfA : pd.DataFrame
        Full table of nuclide predictions for each elevation bin and each erate
        Keys are: 'name', 'nuclide', 'E_cmyr',
                  'NpredSt', 'NpredLm', 'NpredLSDn', # at/g as returned by the server
                  'St_wt', 'Lm_wt', 'LSDn_wt'        # at/g weighted by topostats['wt']
    diagnostics : str
    version : dict
        Returned from the server.
        
    Raises
    -----
    ValueError
        "Cannot get valid input for the online calculator:..."
        "Missing column 'wt' in topostats"
    
    """
    
    if url is None: url = Params.url
    
    if isinstance(erates, Number):
        erates = [erates]
    
    if len(erates)==0:
        raise TypeError("get_NofE_FullTable() argument 'erates' must have at least one value")
        
    if isinstance(sample_data, pd.DataFrame) and len(sample_data)!=1:
        raise TypeError("get_NofE_FullTable() argument 'sample_data' must "+
                         "be dict, pandas Series or single-row pandas Dataframe")
        
    # if topostats is dict or pd.Series cast to DataFrame
    # set 'wt'=1.0 if not defined
    # Note the .map at the end of this function! topostats must be pd.DataFrame with column 'wt'
    if isinstance(topostats, dict):
        temp = pd.DataFrame(topostats, index=[0])
        if 'wt' not in topostats.keys():
            temp['wt'] = 1.
        topostats = temp
        
    if isinstance(topostats, pd.Series): # e.g. if a single row of topostats is passed 
        temp = pd.DataFrame(data=[topostats])
        if 'wt' not in temp.keys():
            temp['wt'] = 1.
        topostats = temp
    
    if 'wt' not in topostats.columns:
        raise ValueError("Missing column 'wt' in topostats")
                         
    # create text_block from all erates and all topotable entries
    tempS = deepcopy(sample_data)
    newline = '.'
    text_block = ''
    for e, erate in enumerate(erates):
        for i, tempT in topostats.iterrows():
            name_str = ("B_{}_{}".format(i,e))
            tempS['name'] = name_str
            tempS['erate'] = erate
            try:
                newline = get_textline(tempS,
                                       tempT, #topostats.iloc[i],
                                       shielding=shielding)
            except (ValueError, TypeError) as e:
                raise ValueError("Cannot get valid input for the online calculator:\n   "+
                                 str(e)) from None
                
            text_block = text_block + newline + ' '

    dfA, diagnostics, version = get_NofE_from_server(text_block, url=url)
    
    # apply weighting factor 'wt' to calculate 'St', 'Lm', 'LSDn'
    if len(dfA)>0:
        # generate a column 'topo' in order to map the weights from dfT.wt to the samples
        dfA[['n', 'topo', 'erate']] = dfA['name'].str.split("_", expand=True)
        dfA.drop(columns=['n', 'erate'], inplace=True)
        
        # apply weighting factor to the exposure ages and get weighted sum of exposure ages
        for x,Nx in zip(Params._scalingmethods, Params._XML_scalingmethods):
            # x='St', 'Nx'='NpredSt' etc:
            dfA[x] = dfA['topo'].astype(int).map(topostats['wt']) * dfA[Nx].astype(float)
        
        num_cols = ['E_cmyr','NpredSt','NpredLm','NpredLSDn', 'St', 'Lm', 'LSDn']
        dfA[num_cols] = dfA[num_cols].apply(pd.to_numeric)
        #dfA['E_cmyr'] = dfA['E_cmyr'].apply(pd.to_numeric)
        dfA.drop(columns=['topo'], inplace=True)
        
    return dfA, diagnostics, version



def get_NofE(
        sample_data:pd.Series,
        topostats:pd.DataFrame,
        shielding:str,
        erates:np.ndarray,
        url:str=None
    ) -> pd.DataFrame:
    """
    Wrapper for get_NofE_FullTable:
    Predicted nuclide concentrations for a given catchment topography
    for a single sample and a suite of erosion rates.
    Note that predicted nuclide conc. are valid for standardization with CF=1
    and can be converted to other standardizations if necessary (the output of
    this function is independent of nuclide concentration and standardization
    specified in 'sample_data').
    
    Note that the index (E_cmyr) may be non-unique.
    This may possibly happens for low erosion rates where the server returns
    NofE for fewer unique erosion rates than requested due to rounding errors;
    It shouldn't happen with the improved string formatting of the textline.
    It happens if erates has duplicates.    
    
    Parameters
    ----------
    sample_data : pd.Series or dict
        Single sample.
    topostats : pd.DataFrame
        Elevation-binned topo data.
    shielding : str or float
        'topo', 'sample' or numeric.
    erates : iterable or float
        (Suite of) erosion rates in cm/yr.
    
    Returns
    ----------
    NofE : pd.DataFrame
        Predicted nuclide concentrations for each erosion rate in erates
        and each scaling method
        Keys are : 'St', 'Lm', 'LSDn'
        Index name is 'E_cmyr'.
    
    Raises
    ------
    ValueError
        Missing or faulty input data
    RuntimeError
        Diagnostics is anything but "No diagnostics".
        
    """
        
    from riversand.utils import validate_topo, validate_nuclide, validate_sample

    if not isinstance(topostats, pd.DataFrame):
        raise TypeError("get_NofE() argument 'topostats' must be pandas DataFrame")

    if isinstance(sample_data, pd.DataFrame):
        if len(sample_data)==1: # recast single-line DataFrame to Series
            sample_data = sample_data.iloc[0]
                    
    if not isinstance(sample_data, (pd.Series, dict)):
        raise TypeError("get_NofE() argument 'sample_data' must be dict, "+
                        "pandas Series or single-row pandas Dataframe")
    
    if isinstance(erates, Number):
        erates = [erates]
    # validate sample_data and a line of topodata to avoid Exceptions raised by
    # get_NofE_FullTable() (ValueError "Cannot get valid input...")
    try:
        _ = validate_topo(topostats.iloc[0])
    except ValueError:
        raise ValueError("Invalid values in 'topostats'") from None
    except IndexError:
        raise ValueError("Empty table 'topostats'") from None
    try:
        _ = validate_sample(sample_data)
    except ValueError:
        raise ValueError("Invalid values in 'sample_data'") from None
    try:
        _ = validate_nuclide(sample_data)
    except ValueError:
        raise ValueError("Invalid values in 'sample_data'") from None
    
    if (shielding=='topo') and ('shielding' not in topostats.keys()):
        raise ValueError("shielding='topo' but no shielding factor defined in 'topostats'")
            
    if (shielding=='sample'):
        try:
            _ = validate_sample(sample_data, shielding=True)
        except ValueError:
            raise ValueError("shielding='sample' but no shielding factor defined in 'sample_data'") from None
         
    dfA, diagnostics, version = get_NofE_FullTable(
        sample_data, topostats, shielding, erates, url=url)
    
    # saturated samples or low nuclide concentration are never an issue with NofE:
    if "No response from" in diagnostics:
        raise RuntimeError("get_NofE() : {}".format(diagnostics))
                        
    if len(dfA)==0: # not sure why this would ever happen
        raise RuntimeError("unexpected error in get_NofE() : empty data table")
    
    # 'sample_name' and 'erosion_rate_cm_yr' are tags returned by the server
    if dfA.groupby('E_cmyr').count().loc[:,'name'].nunique()==len(erates): #1:
        ### if grouping by erosion rate is done correctly each group should have the exact same number of entries
        # if grouping by erosion rate is done correctly there should be as many groups as erosion rates
        NofE = dfA.groupby('E_cmyr').sum(numeric_only=True)
        
    else:
        # rounding errors in the erosion rates sent to the server may lead to problems
        temp = dfA['name'].str.split('_', expand=True)
        dfA['temp'] = temp[2]
        temp = dfA.groupby('temp').sum(numeric_only=True)
        tempE = dfA.groupby('temp').mean(numeric_only=True)
        temp['E_cmyr'] = tempE['E_cmyr']
        NofE = temp.set_index('E_cmyr')
            
    return NofE


def guess_erates(*args, **kwargs) -> np.ndarray:
    """
    Generate a suite of initial erosion rates in cm/yr.
    
    One positional argument yields erosion rates ranging from 0.5*E to 2*E
    > guess_erates(0.001, N=10)
    
    Two positional arguments yields erosion rates ranging from E1 to E2
    > guess_erates(0.001, 0.003, N=5)
    
    Positional arguments can also be a dict of erosion rates with keys
        'St', 'Lm', 'LSDn' if 'scaling' is specified as keyword argument.        
    
    'N' defines the number of output erosion rates; defaults to N=6

    Raises
    ------
    KeyError
    
    """
    
    if 'N' in kwargs.keys(): N = kwargs['N']
    else: N = 6

    if 'scaling' in kwargs.keys(): scaling = kwargs['scaling']
    else: scaling = 'St'
            
    if len(args)==2:
        E1 = args[0]
        E2 = args[1]
        
        if isinstance(E1, Number) and isinstance(E2, Number):
            pass
        elif isinstance(E1, dict) and isinstance(E2, dict):
            try:
                E1 = E1[scaling]
                E2 = E2[scaling]
            except KeyError as e:
                raise e
                #ValueError("guess_erates() invalid keyword argument "+
                #"scaling='{}'".format(scaling))
        else:
            raise KeyError("guess_erates() invalid arguments")
       
    elif len(args)==1:
        E = args[0]
        
        if isinstance(E, Number):
            pass
        elif isinstance(E, dict):
            try:
                E = E[scaling]
            except KeyError as e:
                raise e
                #ValueError("guess_erates() invalid keyword argument "+
                #                 "scaling '{}'"
                #                 .format(scaling))
        else:
            raise KeyError("guess_erates() invalid arguments")
        E1 = E/2
        E2 = 2*E
    else:
        raise KeyError("guess_erates() invalid arguments")
        
    minE = min(E1, E2)
    maxE = max(E1, E2)
    
    minE = np.log(max([minE, 0.001*maxE]))
    maxE = np.log(maxE)
    return np.exp(np.linspace(minE, maxE, int(N))) # cm/yr


def NofE_fitfunc(x, a, b, c):
    return  a/x**2 + b/x + c

def NofE_linfitfunc(x, a, b):
    return  a*x + b


def poly_E_results(
        sample_data:pd.Series,
        topostats:pd.DataFrame,
        shielding:str,
        erates:np.ndarray,
        scaling:str,
        url:str=None,
        strict=True
    ) -> (Number, tuple, pd.DataFrame, list):
    """
    Calculate catchmentwide erosion rate E and uncertainty delE in cm/yr for a
    single sample based on 'topostats' and a suite of initial erosion rates.
    
    Parameters
    ----------
    sample_data : pd.Series or dict
        Single sample. Will be restandardized.
    topostats : pd.DataFrame
        Elevation-binned topo data.
    shielding : str or float
        'topo', 'sample' or numeric.
    erates : iterable or scalar
        (Suite of) erosion rates in cm/yr.
    scaling : str
        Scaling method 'St', 'Lm' or 'LSDn'.
    
    Returns
    ----------
    E : float
        Erosion rate in cm/yr.
    delE : tuple of floats
        Uncertainty (-delE, +delE).
    NofE : pd.Series
        Keys are defined by 'scaling' and index 'E_cmyr'.
    RMSE : float
        Root Mean Squared Error
    error : list of error strings.
        In case of 'minE too high' or 'maxE too low' the list shows only this error.
    
    """

    #from riversand.utils import restandardize_item
    from riversand import utils
    if url is None: url = Params.url
    
    if scaling not in Params._scalingmethods:
        raise ValueError("Invalid scaling '{}' (must be 'St', 'Lm' or 'LSDn')")
    # other requirements are validated by get_NofE() and raised as TypeError or ValueError
    
    if isinstance(erates, Number):
        erates = [erates]
        
    if strict:
        if len(erates)<4:
            raise RuntimeError("poly_E_results() : argument 'erates' should "+
                               "have at least 4 values; use argument "+
                               "'strict=False' to override")
            # note that polynomial fitting requires at least 4 data points
            # 
    # return nan and error string in case of error
    y = None
    E_poly = np.nan
    delE = (np.nan, np.nan)
    RMSE = np.nan
    
    error = []
    
    #erates[:] = np.sort(erates)
    erates = np.sort(list(set(erates))) # drop duplicates
    
    if isinstance(sample_data, pd.DataFrame):
        raise TypeError("poly_E_results() : 'sample_data' must be dict or pandas Series")
        
    #restandardize any dict or pd.Series with columns N, delN, nuclide, standardization
    try:
        sample_data = utils.restandardize(pd.DataFrame([sample_data])).iloc[0]
    except KeyError: # "restandardize() needs columns 'N', ...."
        raise ValueError("invalid sample data")
        
    try:
        NofE = get_NofE(sample_data, topostats, shielding, erates, url=url)
    except TypeError as e: # input data format
        raise e
    except ValueError as e: # input data
        raise e
    except RuntimeError as e: # no server response
        raise e
    
    if not NofE.index.is_unique:
        # rounding errors in the erates sent to the server may result in non-unique indices;
        # drop duplicates
        NofE = NofE[~NofE.index.duplicated(keep='first')]
        error += ["Server cannot resolve erosion rates, dropping duplicates"]
        #raise Warning("Server cannot resolve the provided erosion rates, dropping duplicates")
    
    Nmeas = float(sample_data['N'])
    
    x = NofE.index
    y = NofE[scaling]
    
    if Nmeas >= y.iloc[0]:
        error = ['minE too high']
        #"ERROR: Erosion rate is outside of specified bracket (<{} cm/yr)".format(x[0]))
        
    elif Nmeas <= y.iloc[-1]:
        error = ['maxE too low']
        #"ERROR: Erosion rate is outside of specified bracket (>{} cm/yr)".format(x[-1]))
    
    elif NofE.index.nunique() < 4:
        popt, pcov = optimize.curve_fit(NofE_linfitfunc, x, y)
        a,b = popt
        sol = optimize.root_scalar(NofE_linfitfunc,
                                   args=(a, b-Nmeas),
                                   method='toms748', bracket=[x[0], x[-1]])
        
        if sol.converged:
            E_poly = sol.root            

        else:
            error += ["Root finding did not converge"]
            #"ERROR: No solution found"
        delN = float(sample_data['delN'])
        delE = (np.abs(delN/a), np.abs(delN/a))
        
        RMSE = np.sqrt(np.sum((y - NofE_linfitfunc(NofE.index, *popt))**2)/len(NofE))
        
        error = ["linear fit"]
        
    else:
        popt, pcov = optimize.curve_fit(NofE_fitfunc, x, y)
        a,b,c = popt
        sol = optimize.root_scalar(NofE_fitfunc,
                                   args=(a, b, c-Nmeas),
                                   method='toms748', bracket=[x[0], x[-1]])
            
        if sol.converged:
            E_poly = sol.root            

        else:
            error += ["Root finding did not converge"]
            #"ERROR: No solution found"

        try:
            delE = uncert_on_E(E_poly, y, sample_data)
        except:
            error += ["Cannot compute uncertainty"]
        
        #RMSE = np.sqrt(np.sum((y - NofE_fitfunc(x, *popt))**2)/len(y))
        RMSE = get_RMSE(pd.Series(y, index=x))
        if (RMSE / Nmeas)>1e-3:
            error += ["NRMSE = {:.2e} suggests a poor fit of the polynomial".format(RMSE / Nmeas)]
            #raise Warning("NRMSE = {:.2e} suggests a poor fit of the polynomial!".format(RMSE / Nmeas))
            #print("NRMSE={:.2e}".format(RMSE/Nmeas))
                         
    return E_poly, delE, NofE[scaling], RMSE, error



def uncert_on_E(E:float, NofE:pd.Series, sample_data:pd.Series) -> tuple:
    """ Uncertainty on E from uncertainty on N and curve fit to NofE. """
    
    if isinstance(sample_data, pd.DataFrame):
        if len(sample_data)==1:
            sample_data = deepcopy(sample_data.iloc[0])
        else:
            raise TypeError("uncert_on_E() argument 'sample_data' must be "+
                            "dict, pandas Series or single-row pandas Dataframe")
    
    N1 = sample_data['N']+sample_data['delN']
    N2 = sample_data['N']-sample_data['delN']
    
    x = NofE.index
    popt, pcov = optimize.curve_fit(NofE_fitfunc, x, NofE)
    a,b,c = popt

    # find values x1 and x2 that bracket the erosion rates E1, E2 (E+/-delE)
    x1, y1 = min(NofE.index), max(NofE)
    x2, y2 = max(NofE.index), min(NofE)
    
    while y1<N1: # stepwise 10% decrease of x1
        x1 = 0.9*x1
        y1 = NofE_fitfunc(x1, a, b, c)
    
    while y2>N2: # stepwise 10% increase of x2
        x2 = 1.1*x2
        y2 = NofE_fitfunc(x2, a, b, c)
    
    # find erosion rates corresponding to N+delN and N-delN
    sol = optimize.root_scalar(NofE_fitfunc, args=(a, b, c-N1), method='toms748', bracket=[x1, E])
    if sol.converged:
        E1 = sol.root
        delE1 = E-E1
        
    sol = optimize.root_scalar(NofE_fitfunc, args=(a, b, c-N2), method='toms748', bracket=[E, x2])
    if sol.converged:
        E2 = sol.root
        delE2 = E2-E

    return (delE1, delE2)



def get_RMSE(NofE):   
    """ Root mean squared error. """
    popt, pcov = optimize.curve_fit(NofE_fitfunc, NofE.index, NofE)
    return np.sqrt(np.sum((NofE - NofE_fitfunc(NofE.index, *popt))**2)/len(NofE))


def get_version(url:str=None) -> dict:
    
    if url is None: url = Params.url
    textline = get_textline({'N':1e5, 'delN':1e3}, {'lat':0, 'long':0, 'elevation':0}, shielding=1)
    
    df, diagn, version = get_erates_from_server(textline, url=url)
    
    if "No response from" in diagn:
        raise RuntimeError("get_version() : {}".format(diagn))
        
    return version

def main():
    pass

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

*******************************************************************************
xml.py  :  decoding of the xml server output 

    Copyright (C) 2023  Konstanze Stübner, kstueb@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*******************************************************************************

Conversion of xml-strings returned by the server to
a pandas DataFrame, a diagnostics string and a version dictionary.

> request_vars = {"mlmfile" : "erosion_input_v3",
                  "reportType" : "XML",
                  "resultType" : "long",
                  "summary" : "no",
                  "plotFlag" : "no",
                  "text_block" : textline }
> form_data = urllib.parse.urlencode(request_vars).encode('ascii')
> result = urllib.request.urlopen(url, form_data)
> result_xml = result.read()

ages_from_xml(result_xml)
erates_from_xml(result_xml)

obsolete read_ages(result_xml)
obsolete read_erates(result_xml)
NofE_from_xml(result_xml)

"""

import pandas as pd
import xml.etree.ElementTree as et


def ages_from_xml(result_xml:str, norm=True) -> (pd.DataFrame, str, dict):
    """
    Translate <calcs_v3_age_data> xml string to pandas DataFrame.
    norm=False exludes the normalized nuclide concentrations.

    Returns
    -------
    df : pandas DataFrame with one row per sample and nuclide.
        Units are years.
    diagnostics : str
        diagnostics string.
    version : dict
        version info returned from the server.
    """
    
    """
    xml (v 2024-05-10)
    root : <calcs_v3_age_data>
    children :
        <exposureAgeResult>   # one for each sample site
        <diagnostics>
        <version>
        
    <exposureAgeResult> :
        item <sample_name>   # one for each sample site
        + following 18 items; sequence repeats if more than one analysis per sample site
        item <t10quartz_St>
        item <delt10quartz_int_St>
        item <delt10quartz_ext_St>
        item <Nnorm10quartz_St>
        item <delNnorm10quartz_int_St>
        item <delNnorm10quartz_ext_St>
        item <t10quartz_Lm>
        item <delt10quartz_int_Lm>
        item <delt10quartz_ext_Lm>
        item <Nnorm10quartz_Lm>
        item <delNnorm10quartz_int_Lm>
        item <delNnorm10quartz_ext_Lm>
        item <t10quartz_LSDn>
        item <delt10quartz_int_LSDn>
        item <delt10quartz_ext_LSDn>
        item <Nnorm10quartz_LSDn>
        item <delNnorm10quartz_int_LSDn>
        item <delNnorm10quartz_ext_LSDn>
    """
    
    version = {}
    # tags returned by the server
    Be10cols = ['t10quartz_St', 'delt10quartz_int_St', 'delt10quartz_ext_St',
                'Nnorm10quartz_St', 'delNnorm10quartz_int_St', 'delNnorm10quartz_ext_St',
                't10quartz_Lm', 'delt10quartz_int_Lm', 'delt10quartz_ext_Lm',
                'Nnorm10quartz_Lm', 'delNnorm10quartz_int_Lm', 'delNnorm10quartz_ext_Lm',
                't10quartz_LSDn', 'delt10quartz_int_LSDn', 'delt10quartz_ext_LSDn',
                'Nnorm10quartz_LSDn', 'delNnorm10quartz_int_LSDn', 'delNnorm10quartz_ext_LSDn']
    
    # empty dataframe formatted to match the output in case of an exception
    out = pd.DataFrame(columns=['sample_name', 'nuclide']+[c.replace('10quartz', '') for c in Be10cols])
    out = out.rename(columns={'sample_name': 'sample'})

    if len(result_xml)==0:
        ret_str = "Empty string returned from the server. This may happen if at least one sample has an extremely low nuclide concentration."
        return out, ret_str, version

    try:
        root = et.fromstring(result_xml)
    except Exception as err:
        print(err)
        return out, err, version

    tags = [i.tag for i in root] # ['exposureAgeResult', 'exposureAgeResult', ..., 'diagnostics', 'version']
    # note that a tag 'exposureAgeResult' may have more than one result

    iD = [i for i, tag in enumerate(tags) if tag=="diagnostics"]
    iV = [i for i, tag in enumerate(tags) if tag=="version"]
    assert len(iD)==1 # exactly 1 element <diagnostics>
    assert len(iV)<=1 # max 1 element <version>; for erosion rates this may be 0 for invalid data

    # version dict:
    if len(iV)==1:
        versionElement = root[iV[0]]
        for item in versionElement:
            version[item.tag] = item.text
            
    # diagnostics:
    dignosticsElement = root[iD[0]]
    diagnostics = dignosticsElement.text

    if diagnostics=="No diagnostics":
        pass
    elif 'saturated' in diagnostics:
        pass
    else:
        # This is a bug in the server output: if there are any issues with
        # the data the server returns some diagnostics but no version info,
        # and the root tag is "exposureAgeResult" even if erosion rate was
        # requested. The data is all jumbled zeroes
        pass
        return out, diagnostics, version
        
    ## START xml conversion
    nuclide_dict = {'10quartz': 'Be-10', '26quartz': 'Al-26'}
    df = pd.DataFrame(columns=['sample_name', 'nuclide']) # tag 'sample_name' returned by the server

    i = -1 # index in result DataFrame
    for element in root: # iterate over 'exposureAgeResult's; each may have more than one analysis
        if element.tag=='exposureAgeResult': # skip the 'dignostics' and 'version' elements
            for item in element:
                #assert item.tag in {'sample_name', 't10quartz_St', 'delt10quartz_int_St', etc}
                
                if item.tag=='sample_name':
                    sample_name = item.text # name for one or several analyses
                    
                elif item.tag in {'t10quartz_St', 't26quartz_St'}:
                    i += 1 # new analysis i.e. next line in results
                    nuclide = item.tag[1:9] # nuclide string '10quartz', '26quartz'
                    df.loc[i, 'sample_name'] = sample_name
                    df.loc[i, 'nuclide'] = nuclide_dict[nuclide] # translate to 'Be-10', 'Al-26'
                    col = item.tag.replace(nuclide, '') # remove the 'xxquartz' substring
                    df.loc[i, col] = item.text
                else:
                    col = item.tag.replace(nuclide, '') # remove the 'xxquartz' substring
                    df.loc[i, col] = item.text
    
    df = df.rename(columns={'sample_name': 'sample'})
    
    #df.nuclide = df.nuclide.map(pd.Series(repl)) # replace '10quartz' by 'Be-10'
    if norm:
        return df, diagnostics, version
    else: # short table excluding the normalized values
        return df.iloc[:,[0,1,2,3,4,8,9,10,14,15,16]], diagnostics, version
    
    
def erates_from_xml(result_xml:str) -> (pd.DataFrame, str, dict):
    """
    Tanslate <calcs_v3_erosion_data> xml string to pandas DataFrame.
    
    Returns
    -------
    df : pandas DataFrame with one row per sample and nuclide.
        Units are density: g/cm3, erosion rate: g/cm2/yr.
        Note that the m/Myr output of the browser version is not part of the
        xml string and can be calculated from the density.
    diagnostics : str
        diagnostics string.
    version : dict
        version info returned from the server.
    """
    
    """
    xml (v 2024-05-10)
    root: <calcs_v3_erosion_data>
    children :
        <erosionRateResult>   # one for each sample site
        <diagnostics>
        <version>
    
    <erosionRateResult> :
        item <sample_name>   # one for each sample site
        item <sample_density>   # one for each sample site
        item <nuclide_result>   # one or several if more than one analysis per sample site
            item <nuclide>
            item <E_gcm2_St>
            item <delE_gcm2_St> # int error
            item <delE_gcm2_St> # ext error
            item <E_gcm2_Lm>
            item <delE_gcm2_Lm>
            item <delE_gcm2_Lm>
            item <E_gcm2_LSDn>
            item <delE_gcm2_LSDn>
            item <delE_gcm2_LSDn>
    """
    
    version = {}
    # tags returned by the server are intentical for int and ext errors!
    cols = ['sample_name', 'nuclide', 'sample_density', 
            'E_gcm2_St',   'delE_int_gcm2_St',   'delE_ext_gcm2_St',
            'E_gcm2_Lm',   'delE_int_gcm2_Lm',   'delE_ext_gcm2_Lm',
            'E_gcm2_LSDn', 'delE_int_gcm2_LSDn', 'delE_ext_gcm2_LSDn']
    
    # empty dataframe formatted to match the output in case of an exception
    out = pd.DataFrame(columns=cols)
    out = out.rename(columns={
        'sample_name': 'sample', 'sample_density': 'density'})
    
    if len(result_xml)==0:
        ret_str = "Empty string returned from the server. This may happen if at least one sample has an extremely low nuclide concentration."
        return out, ret_str, version
    
    try:
        root = et.fromstring(result_xml)
    except Exception as err:
        print(err)
        return out, err, version

    tags = [i.tag for i in root] # ['erosionRateResult', 'erosionRateResult', ..., 'diagnostics', 'version']
    # note that 'erosionRateResult' may have more than one result
    
    iD = [i for i, tag in enumerate(tags) if tag=="diagnostics"]
    iV = [i for i, tag in enumerate(tags) if tag=="version"]
    assert len(iD)==1 # exactly 1 element <diagnostics>
    assert len(iV)<=1 # max 1 element <version>; may be 0 for invalid data
    
    # version dict:
    if len(iV)==1:
        versionElement = root[iV[0]]
        for item in versionElement:
            version[item.tag] = item.text
            
    # diagnostics:
    dignosticsElement = root[iD[0]]
    diagnostics = dignosticsElement.text
    
    if diagnostics=="No diagnostics":
        pass
    elif 'saturated' in diagnostics:
        pass
    else:
        # This is a bug in the server output: if there are any issues with
        # the data the server returns some diagnostics but no version info,
        # and the root tag is "exposureAgeResult" even if erosion rate was
        # requested. The data is all jumbled zeroes
        return out, diagnostics, version
        
    ## START xml conversion
    nuclide_dict = {'N10quartz': 'Be-10', 'N26quartz': 'Al-26'}    
    df = pd.DataFrame(columns=cols) # these are the tags actually returned from the server
    
    i = -1
    for element in root: # iterate over 'erosionRateResult's; each may have more than one analysis
        if element.tag=='erosionRateResult': # skip the 'dignostics' and 'version' elements
            sample_name = None
            sample_density = None
            
            for item in element:
                assert item.tag in {'sample_name', 'sample_density', 'nuclide_result'}
                
                if item.tag=='sample_name':
                    sample_name = item.text # name for one or several analyses
                
                if item.tag=='sample_density':
                    sample_density = item.text
                
                if item.tag=='nuclide_result':
                    i += 1 # new analysis i.e. next line in results
                    
                    df.loc[i, 'sample_name'] = sample_name
                    df.loc[i, 'sample_density'] = sample_density
                
                    # sort out internal and external errors with identical item tags
                    delInt = True # errors are reported first int then ext
                    for itm in item:
                        
                        if 'delE' in itm.tag:
                            if delInt==True:
                                tag = itm.tag[:4]+"_int"+itm.tag[4:]
                                delInt = False
                            else:
                                tag = itm.tag[:4]+"_ext"+itm.tag[4:]
                                delInt = True
                                
                            df.loc[i, tag] = itm.text # <delE_gcm2_XXX>
                        
                        elif itm.tag=='nuclide':        
                            df.loc[i, 'nuclide'] = nuclide_dict[itm.text] # translate to 'Be-10', 'Al-26'
                        
                        else:
                            df.loc[i, itm.tag] = itm.text # <E_gcm2_XX>
                            
                    
    df = df.rename(columns={
        'sample_name': 'sample', 'sample_density': 'density'})
    
    return df, diagnostics, version

    

def NofE_from_xml(result_xml:str) -> (pd.DataFrame, str, dict):
    """    
    Returns
    -------
    df : pandas DataFrame with one row per sample and nuclide.
        Units are erosion rate: g/cm2/yr, nuclide concentrations atoms/gram
    diagnostics : str
        diagnostics string.
    version : dict
        version info returned from the server.
    """
    
    version = {}
    cols = ['sample_name', 'nuclide', 'erosion_rate_cm_yr',
            'NpredSt', 'NpredLm', 'NpredLSDn']
    df = pd.DataFrame(columns=cols)
    
    out = df.rename(columns={
        'sample_name' : 'name', 'erosion_rate_cm_yr': 'E_cmyr',
        })
    
    if len(result_xml)==0:
        ret_str = "Empty string returned from the server; this may happen for a very small nuclide concentration"
        return out, ret_str, {}
    
    try:
        root = et.fromstring(result_xml)
    except Exception as err:
        print(err)
        return out, err, {}

    tags = [i.tag for i in root]
    
    iD = [i for i, tag in enumerate(tags) if tag=="diagnostics"]
    iV = [i for i, tag in enumerate(tags) if tag=="version"]
    assert len(iD)==1 # exactly 1 Element <diagnostics>
    assert len(iV)<=1 # max 1 Element <version>; may be 0 for invalid data

    iE = [i for i, tag in enumerate(tags) if tag=="NofEResult"]
    
    # diagnostics:
    if len(iD)==1:
        dignosticsElement = root[iD[0]]
        diagnostics = dignosticsElement.text
        
        if diagnostics=="No diagnostics":
            pass
        elif 'saturated' in diagnostics:
            pass
        else:
            # This is a bug in the server output: if there are any issues with
            # the data the server returns some diagnostics but no version info,
            # and the root tag is "exposureAgeResult" even if erosion rate was
            # requested. The data is all jumbled zeroes
            return out, diagnostics, {}
        
    # version dict:
    if len(iV)==1:
        versionElement = root[iV[0]]
        for item in versionElement:
            version[item.tag] = item.text
   
    i = 0
    for NofEResult in [root[i] for i in iE]:
        assert NofEResult.tag=='NofEResult'
        sample_name = None
        erosion_rate_cm_yr = None
        
        for item in NofEResult:
            assert item.tag in {'sample_name', 'erosion_rate_cm_yr', 'nuclide_result'}
            
            if item.tag=='sample_name':
                sample_name = item.text
            
            if item.tag=='erosion_rate_cm_yr':
                erosion_rate_cm_yr = item.text
            
            if item.tag=='nuclide_result':
                this_nuclide = {'sample_name': sample_name,
                                'erosion_rate_cm_yr': erosion_rate_cm_yr
                               }
                for E in item:
                    this_nuclide[E.tag] = E.text # <nuclide_result> to dict
                        
                df.loc[i, :] = this_nuclide
                i += 1
                    
    df.rename(columns={
        'sample_name' : 'name', 'erosion_rate_cm_yr': 'E_cmyr',
        },
        inplace=True)
    
    return df, diagnostics, version


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

*******************************************************************************
params.py  :  default parameters

    Copyright (C) 2023  Konstanze Stübner, kstueb@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*******************************************************************************

"""
import os

class Params:
    out_path = './plots'

    #url = "https://hess.ess.washington.edu/cgi-bin/matweb"
    url = "http://stoneage.hzdr.de/cgi/matweb"
    
    units = {'cm/yr': 1, 'cm/kyr': 1000, 'mm/yr': 10, 'mm/kyr': 10000, 'm/Myr': 10,
              'cm/a': 1, 'cm/ka': 1000, 'mm/a': 10, 'mm/ka': 10000, 'm/Ma': 10,
              }
    
    # # see http://hess.ess.washington.edu/math/docs/v3/v3_input_explained.html
    default_values = {'name'       : 'Test',
                      'press_flag' : 'std',    # 'std', 'ant'
                      'thickness'  : 0,        # in cm
                      'density'    : 2.65,     # in g/cm3
                      'shielding'  : 1,        # 0 to 1
                      'erate'      : 0,        # in cm/yr
                      'year'       : 2010,
                      'nuclide'    : 'Be-10',  # 'Be-10', 'Al-26'
                      'mineral'    : 'quartz', # 'quartz'
                      }
    
    _default_standards = {'Be-10': '07KNSTD',
                          'Al-26': 'KNSTD',
                         }
    



    Be_stds = {"07KNSTD":1.0000,"KNSTD":0.9042,"NIST_Certified": 1.0425,
               "LLNL31000":0.8761,"LLNL10000":0.9042,"LLNL3000": 0.8644,
               "LLNL1000":0.9313,"LLNL300":0.8562,"NIST_30000": 0.9313,
               "NIST_30200":0.9251,"NIST_30300":0.9221,"NIST_30600": 0.9130,
               "NIST_27900":1.0000,"S555":0.9124,"S2007":0.9124,
               "BEST433":0.9124,"BEST433N":1.0000,"S555N":1.0000,
               "S2007N":1.0000,"STD11":1.0000,"0":0,"NIST_30500":0.9124,
               "SMDBe12": 1.0000}
    Al_stds = {"KNSTD":1.0000,"ZAL94":0.9134,"SMAL11":1.021,"0":0,
               "ZAL94N":1.0000,"ASTER":1.021,"Z92-0222":1.0000,
               "SMDAl11":1.018}
    
    _scalingmethods = ['St', 'Lm', 'LSDn']
    
    _XML_scalingmethods = ['NpredSt', 'NpredLm', 'NpredLSDn'] # returned by the online calculator
    
    # keys for the display of sample data
    all_keys = ['name',
                'lat', 'latitude', 'long', 'longitude', 'elev', 'elevation',
                'press_flag', 'thickness', 'density', 'shielding', 'erate', 'year',
                'nuclide', 'mineral', 'N', 'delN', 'standardization']
            
    @classmethod
    def set_outpath(cls, value):
        try:
            os.makedirs(value)
        except FileExistsError:
            cls.out_path = value
        except PermissionError:
            print("No permission to make directory '{}'; saving plots to '{}'."
                  .format(value, cls.out_path))
        else:
            cls.out_path = value


def set_folder(value):
    Params.set_outpath(value)

def get_folder():
    print("Plots are saved in folder '{}'.".format(Params.out_path))

def print_units():
    print("Valid units of erosion rate are: {}.".format(', '.join(Params.units.keys())))
    
def print_defaults():
    print("\nDefault values:")
    for k,v in Params.default_values.items():
        print("  {0: <10} : {1: <10}".format(k, v))
        
    print("\nDefault standardizations:")
    for k,v in Params._default_standards.items():
        print("  {0: <10} : {1: <10}".format(k, v))
    print()
    
def print_standardizations():
    print("\nBe-10 standardizations:")
    for k,v in Params.Be_stds.items():
        print("  {0: <10} : {1: <10}".format(k, v))
    print("\nAl-26 standardizations:")
    for k,v in Params.Al_stds.items():
        print("  {0: <10} : {1: <10}".format(k, v))
    print()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

*******************************************************************************
utils.py  :  utilities for the riversand package 

    Copyright (C) 2023  Konstanze Stübner, kstueb@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*******************************************************************************

Validation functions for online calculator:
    validate_topo()
    validate_nuclide()
    validate_sample()
    - Return validated and complete dictionaries
    - Fill in default values from params.Params if non-mandatory keys are missing
    - Raise ValueError if values are invalid / mandatory values are missing 


restandardize()

    
Geospatial processing functions:
    feature_in_raster()
    clip_raster()
    get_xarray_centroid()
    get_polygon_centroid()
    projected_xy_to_longlat()
    get_bins()
    get_topostats()
    

import_data()

"""

import numpy as np
import pandas as pd
pd.options.mode.copy_on_write = True

import re
from numbers import Number
from copy import deepcopy

import rasterio
import rasterio.crs
import rasterio.mask
from rasterio import MemoryFile

import xarray as xr

from riversand.params import Params

# =============================================================================
# Validation for online calculator
# =============================================================================

def valid_namestring(name:str) -> bool:
    name_regex = "^[a-zA-Z0-9_-]*$"
    if len(name)>32:
        return False
    if not re.match(name_regex, name):
        return False
    return True


def validate_topo(item, shielding:bool=False) -> dict:
    """
    Generate dict of topo data.
    
    Parameters
    ----------
    item : dict or pd.Series
        Single sample.
        Mandatory keys 'lat', 'long', 'elevation'. Optional key 'shielding'.
    
    Raises
    ------
    ValueError
        Mandatory item missing or ill-defined.
    
    """
           
    out = dict()
    
    # mandatory keys:
    if ('lat' not in item.keys() or
        item['lat'] is None):
        raise ValueError("Latitude is not defined")
        
    if ('long' not in item.keys() or
        item['long'] is None):
        raise ValueError("Longitude is not defined")
        
    if ('elevation' not in item.keys() or
        item['elevation'] is None):
        raise ValueError("Elevation is not defined")
        
    try:
        out['lat'] = float(item['lat'])
    except: # ValueError (e.g. str) / TypeError (e.g. list)
        out['lat'] = np.nan
    if np.isnan(out['lat']):
        raise ValueError("Latitude is not a number")
    if not (-90 <= out['lat'] <= 90):
        raise ValueError("Latitude is out of bounds (-90..+90)")
    
    try:
        out['long'] = float(item['long'])
    except: # ValueError (e.g. str) / TypeError (e.g. list)
        out['long'] = np.nan
    if np.isnan(out['long']):
        raise ValueError("Longitude is not a number")
    if not (-180 <= out['long'] <= 180):
        raise ValueError("Longitude is out of bounds (-180..+180)")
    
    try:
        out['elevation'] = float(item['elevation'])
    except: # ValueError (e.g. str) / TypeError (e.g. list)
        out['elevation'] = np.nan
    if np.isnan(out['elevation']):
        raise ValueError("Elevation is not a number")
    if out['elevation'] < (-500):
        raise ValueError("Elevation is less than -500")
        
    # optional keys:
    if shielding:
        if ('shielding' not in item.keys() or
            item['shielding'] is None):
            raise ValueError("Shielding is not defined")
        try:
            out['shielding'] = float(item['shielding'])
        except: # ValueError (e.g. str) / TypeError (e.g. list)
            out['shielding'] = np.nan
        if np.isnan(out['shielding']):
            raise ValueError("Shielding is not a number")
        if not (0 <= out['shielding'] <= 1):
            raise ValueError("Shielding is out of bounds (0..1)")
            
    return out


def validate_nuclide(item) -> dict:
    """
    Generate dict of nuclide data.

    Parameters
    ----------
    item : dict or pd.Series
        Single sample.
        Mandatory keys 'N', 'delN'. Default values for 'name', 'nuclide',
        'mineral', 'standardization'.
    
    Raises
    ------
    ValueError
        Mandatory item missing or ill-defined.
    
    """
        
    out = dict()
    
    # mandatory keys:
    if ('N' not in item.keys() or
        item['N'] is None):
        raise ValueError("Concentration N is not defined")
        
    if ('delN' not in item.keys() or
        item['delN'] is None):
        raise ValueError("Uncertainty delN is not defined")
        
        
    try:
        out['N'] = float(item['N']) # accepts string "1.2E+6"
    except: # ValueError (e.g. str) / TypeError (e.g. list)
        raise ValueError("Concentration N is not a number") from None
    if not out['N'] > 0:
        raise ValueError("Concentration N is 0 or less")
        
    try:
        out['delN'] = float(item['delN'])
    except: # ValueError (e.g. str) / TypeError (e.g. list)
        raise ValueError("Uncertainty delN is not a number") from None
    if not out['delN'] > 0:
        raise ValueError("Uncertainty delN is 0 or less")
        
    # optional keys with default values:
    if 'name' in item.keys():
        #name_regex = "^[a-zA-Z0-9_-]*$"
        try:
            out['name'] = str(item['name'])
        except:
            out['name'] = '??' # illegal character
        
        if not valid_namestring(out['name']):
            raise ValueError("Illegal characters in sample name or >32 characters")
    else:
        out['name'] = Params.default_values['name'] # default value
    
    if 'nuclide' in item.keys():
        try:
            out['nuclide'] = str(item['nuclide'])
        except:
            out['nuclide'] = '??'
        if out['nuclide'] not in {'Be-10', 'Al-26'}:
            raise ValueError("Illegal nuclide")
    else:
        out['nuclide'] =Params.default_values['nuclide'] # default value
        if 'standardization' in item.keys():
            raise ValueError("Must specify nuclide if standardization is specified")
        
    if 'mineral' in item.keys():
        try:
            out['mineral'] = str(item['mineral'])
        except:
            out['mineral'] = '??'
        if out['mineral'] != 'quartz':
            raise ValueError("Illegal mineral")
    else:
        out['mineral'] = Params.default_values['mineral'] # default value

    if out['nuclide'] == 'Be-10':
        if 'standardization' in item.keys():
            try:
                out['standardization'] = str(item['standardization'])
            except:
                out['standardization'] = '??'
        else:
            out['standardization'] = Params._default_standards[out['nuclide']]
        if out['standardization'] not in Params.Be_stds.keys():
            raise ValueError("Illegal Be standardization")

    if out['nuclide'] == 'Al-26':
        if 'standardization' in item.keys():
            try:
                out['standardization'] = str(item['standardization'])
            except:
                out['standardization'] = '??'
        else:
            out['standardization'] = Params._default_standards[out['nuclide']]
        if out['standardization'] not in Params.Al_stds.keys():
            raise ValueError("Illegal Al standardization")    

    return out


def validate_sample(item, shielding:bool=False) -> dict:
    """
    Generate dict of sample data.
    
    Parameters
    ----------
    item : dict or pd.Series
        Single sample.
        Default values for 'press_flag', 'density', 'thickness', 'erate',
        'year'. Optional key 'shielding' .
    
    Raises
    ------
    ValueError
        Mandatory item missing or ill-defined.
    
    """
        
    out = dict()

    # optional keys with default values:
    if 'press_flag' in item.keys():
        try:
            out['press_flag'] = str(item['press_flag'])
        except:
            out['press_flag'] = '??'
        if out['press_flag'] not in {'std','ant'}:
            raise ValueError("Illegal pressure flag")
    else:
        out['press_flag'] = Params.default_values['press_flag'] # default value
            
    if 'density' in item.keys():
        try:
            out['density'] = float(item['density'])
        except:
            out['density'] = np.nan
        if np.isnan(out['density']):
            raise ValueError("Density is not a number")
        if out['density'] <= 0:
            raise ValueError("Density is 0 or less")
    else:
        out['density'] = Params.default_values['density'] # default value
        
    if 'thickness' in item.keys():
        try:
            out['thickness'] = float(item['thickness'])
        except:
            out['thickness'] = np.nan
        if np.isnan(out['thickness']):
            raise ValueError("Thickness is not a number")
        if out['thickness'] < 0:
            raise ValueError("Thickness is less than 0")
    else:
        out['thickness'] = Params.default_values['thickness'] # default value
        
    if 'erate' in item.keys():
        try:
            out['erate'] = float(item['erate'])
        except:
            out['erate'] = np.nan
        if np.isnan(out['erate']):
            raise ValueError("Erosion rate is not a number")
        if out['erate'] < 0:
            raise ValueError("Erosion rate is less than 0")
    else:
        out['erate'] = Params.default_values['erate'] # default value

    if 'year' in item.keys():
        try:
            out['year'] = round(float(item['year']))
        except:
            raise ValueError("Collection year is not a number") from None
        if not (1700 <= out['year'] <= 2100):
            raise ValueError("Collection year is out of bounds (1700..2100)")
    else:
        out['year'] = Params.default_values['year'] # default value
        
    # optional keys:
    if shielding:
        if ('shielding' not in item.keys() or
            item['shielding'] is None):
            raise ValueError("Shielding is not defined")
        try:
            out['shielding'] = float(item['shielding'])
        except:
            out['shielding'] = np.nan
        if np.isnan(out['shielding']):
            raise ValueError("Shielding is not a number")
        if not (0 <= out['shielding'] <= 1):
            raise ValueError("Shielding is out of bounds (0..1)")

    return out


def restandardize(df) -> pd.DataFrame:
    """
    Converts any pd.DataFrame with columns N, delN, nuclide, standardization
    to default standardization.
     
    Raises
    ------
    KeyError
        Mandatory column missing
        
    """

    if not all([k in df.columns for k in {'N', 'delN', 'nuclide', 'standardization'}]):
        raise KeyError("restandardize() needs columns 'N', 'delN', 'nuclide', 'standardization'. "+
                       "Try using import_data() to recast data to required format.")
    
    df['_N'], df['_delN'], df['_standardization'] = df['N'], df['delN'], df['standardization']
    
    df['N'], df['delN'], df['standardization'] = np.nan, np.nan, ''
    
    nucls = ['Be-10', 'Al-26']
    stdss = [pd.Series(Params.Be_stds), pd.Series(Params.Al_stds)]
    for nu, st in zip(nucls, stdss): 
        sel = df['nuclide']==nu
        
        df.loc[sel, 'N'] = (
            df.loc[sel,'_standardization'].map(st)
            * df.loc[sel,'_N']
            )
        df.loc[sel, 'delN'] = (
            df.loc[sel,'_standardization'].map(st)
            * df.loc[sel,'_delN']
            )
        df.loc[sel*df['N'].notnull(), 'standardization'] = (
            Params._default_standards[nu]
            )
    df.drop(columns=['_N', '_delN', '_standardization'], inplace=True) 
    return df



# =============================================================================
# Geospatial processing
# =============================================================================
    
def feature_in_raster(polygon:dict, src:rasterio.DatasetReader) -> bool:
    """
    Check whether feature is within the bounds of a raster.
    
    Parameters
    ----------
    polygon : catchment.catchments[0]['geometry']
    
    """
    
    if 'coordinates' not in polygon.keys():
        raise TypeError("feature_in_raster() argument 'polygon' not a valid polygon dictionary")
    if not isinstance(src, rasterio.io.DatasetReader):
        raise TypeError("feature_in_raster() argument 'src' not a valid rasterio.DatasetReader")

    (left, bott, right, top) = rasterio.features.bounds(polygon)
    bounds = src.bounds
    return ((bounds[0] < left) and (bounds[1] < bott) and
            (bounds[2] > right) and (bounds[3] > top))


def clip_raster(polygon:dict, src:rasterio.DatasetReader,
                label:str,
                name='') -> xr.DataArray:
    """
    Clip raster with catchment polygon and return as xr.DataArray
    with attributes 'transform', 'label' and 'name'.
    > clip['elevation'].attrs['name']
    
    Parameters
    ----------
    polygon : dict
        > polygon = rv.catchment.catchments[0]['geometry']
    label : str
        'elevation', 'shielding', 'quartz'
    name : str
        catchment name
        > rv.catchments.catchments[0].properties[rv.cid]
    
    Raises
    ------
    TypeError
    ValueError
        "clip_raster() : catchment polygon out of bounds"
        
    """
    
    if 'coordinates' not in polygon.keys():
        raise TypeError("clip_raster() argument 'polygon' not a valid polygon dictionary")
    if not isinstance(src, rasterio.io.DatasetReader):
        raise TypeError("clip_raster() argument 'src' not a valid rasterio.DatasetReader")

    if src.closed:
        src = rasterio.open(src.name, 'r')

    if not feature_in_raster(polygon, src):
        raise ValueError("clip_raster() : catchment polygon out of bounds")

    # read the raster data with rasterio and clip the catchment
    try:
        Z, Z_transform = rasterio.mask.mask(src, [polygon], nodata=np.nan,
                                            crop=True)
        
    except TypeError: # recast raster to float to facilitate nodata=nan
        data = src.read()
        profile = src.profile
        src.close()

        profile.update(dtype='float32', driver='GTiff')
        with MemoryFile() as memfile:
            with memfile.open(**profile) as dst: # open as DatasetWriter
                dst.write(data)
                del data

            src = memfile.open()  # reopen as DatasetReader
            
        Z, Z_transform = rasterio.mask.mask(src, [polygon], nodata=np.nan,
                                            crop=True)
        
    Z_meta = src.meta
    #Z_bounds = src.bounds
    Z_meta.update({'driver': 'GTiff',
           'height': Z.shape[1],
           'width': Z.shape[2],
           'transform': Z_transform})
    
    # Convert raster to xr.DataArray
    xx = (np.arange(Z_meta['width'])*Z_transform[0] +
          Z_transform[2] + 0.5*Z_transform[0])
    yy = (np.arange(Z_meta['height'])*Z_transform[4] +
          Z_transform[5] + 0.5*Z_transform[4])
    Z = xr.DataArray(Z.squeeze(), dims=('y','x'), coords={
                              'x' : xx,
                              'y' : yy },
                    )
    src.close()
    
    Z.attrs['transform'] = Z_transform
    Z.attrs['label'] = label
    Z.attrs['name'] = name
    
    return Z


def get_xarray_centroid(X:xr.DataArray) -> tuple:
    """
    Parameters
    ----------
    X : xr.DataArray
        Projected raster dataset clipped to the catchment polygon.
        > X = clips['elevation']
    
    Returns
    -------
    (centr_x, centr_y) : tuple of floats
        Projected longitude, latitude. 
    
    Raises
    ------
    RuntimeError
        "empty xarray" if no values in X>0 (may be caused by quartz correction).
        
    """

    A = xr.DataArray((X.values>0), coords=X.coords)
    if sum(sum(A))==0:
        raise RuntimeError('empty xarray')
    centr_x = float(np.sum(A.sum('y')/np.sum(A.sum('y')) * A.x))
    centr_y = float(np.sum(A.sum('x')/np.sum(A.sum('x')) * A.y))
    return (centr_x, centr_y)

    

def get_polygon_centroid(polygon:dict) -> tuple:
    """
    Parameters
    ----------
    polygon : dict
        > polygon = catchment.catchments[0]['geometry']
    
    Returns
    -------
    (centr_x, centr_y) : tuple of floats
        Projected longitude, latitude.
    
    """
    
    if 'coordinates' not in polygon.keys():
        raise TypeError("argument 'polygon' not a valid polygon dictionary")
        
    from shapely.geometry import Polygon
        
    centr_x = Polygon(polygon['coordinates'][0]).centroid.x
    centr_y = Polygon(polygon['coordinates'][0]).centroid.y

    return (centr_x, centr_y)

    
def projected_xy_to_longlat(xy:tuple, epsg:int) -> tuple:
    """  
    Parameters
    ----------
    xy : tuple
        Projected coordinates
    epsg : int
        epsg code
    
    Returns
    -------
    (long, lat) : tuple of floats
        Unprojected longitude, latitude.
    
    """
    
    if np.nan in xy:
        return (np.nan, np.nan)

    from pyproj import Transformer
    # note that 4326 is lat/lon, whereas utm is x/y
    # use option always_xy=True to avoid trouble
    transformer = Transformer.from_crs(epsg, 4326, always_xy=True)
    #transformer = Transformer.from_crs("EPSG:4326", "EPSG:26917")

    (long, lat) =  transformer.transform(xy[0], xy[1])
    return (long, lat)


def get_bins(Z:xr.DataArray, binsize:int=100) -> np.ndarray:
    """
    Get pretty elevation bins.
    
    Parameters
    ----------
    Z : ndarray or xr.DataArray
        Values to be binned.
    
    """

    return np.arange(start=np.floor(Z.min()/binsize)*binsize,
                     step=binsize,
                     stop=np.ceil(Z.max()/binsize)*binsize + binsize,
                     dtype=int
                     )


def get_topostats(clips, bins, centroid='from_clipped',
                  polygon=None,
                  epsg=None,
                  validate_transforms=True,
                  ) -> (pd.DataFrame, dict):
    """
    Compute elevation statistics. 
    
    Parameters
    ----------
    clips : dict of xr.DataArrays
        Keys are raster dtypes ('elevation', 'shielding, ...).
        Values have attributes .transform (Affine) and .label (dtype).
        Additional key 'epsg' has the projection.
        > clips = rv.clip_all_rasters(n)
            
    centroid : str or tuple
        "from_clipped" or (long, lat) in projected coordinates.
    
    Returns
    -------
    df : pd.DataFrame
        Topo statistics.
    summary : dict
        Percentile elevations (35, 50, 65%) and centroid coordinates.
        Keys are 'elevLo', 'elev50', 'elevHi', 'lat', 'long', 'areakm2'.
        
    Raises
    ------
    ValueError
        "epsg cannot be determined from 'clips'; specify as keyword argument"
        "argument 'clips' with non-matching shape/resolution"
        "argument 'clips' with non-matching transforms"
    TypeError
        "centroid must be string 'from_clipped' or tuple (long, lat)"
    """

    from riversand.geospatial import Raster
    
    if isinstance(epsg, int):
        pass
    elif 'epsg' in clips.keys(): # clips['epsg']=None for non-validated 
        epsg = clips['epsg']
    if epsg is None:
        raise ValueError("epsg cannot be determined from 'clips'; "+
                         "specify as keyword argument")
    
    if isinstance(centroid, str): # allow mistakes
        centroid = centroid.lower()
        if centroid in 'from_clipped':
            centroid = 'from_clipped'            
    elif (isinstance(centroid, tuple) and len(centroid)==2):
        pass
    else:
        raise TypeError("centroid must be string 'from_clipped' "+
                        "or tuple (long, lat)")
    
    # Z: xr.DataArray of elevation raster
    Z = clips['elevation']

    # XX: dict of valid raster datasets; specifically excluding clips['epsg'] and clips['name']
    XX = dict([(key,val) for key,val in clips.items() if key in Raster.dtypes])
    
    # Perform some validation of the xr.DataArrays.
    # should be redundant if clips = clip_all_rasters()
    if len(XX)>1:
        shape = Z.shape #clips['elevation']
        
        for X in XX.values(): 
            if not X.shape==shape:
                raise ValueError("argument 'clips' with non-matching shape/resolution")

        if validate_transforms:
            transform = clips['elevation'].transform
            for X in XX.values():
                if not X.transform==transform:
                    raise ValueError("argument 'clips' with non-matching transforms")

    
    # Compute bins and centroid coordinates
    if isinstance(bins, Number):
        bins = get_bins(Z, bins)
    
    if centroid=='from_clipped':
        try:
            (centr_x, centr_y) = get_xarray_centroid(Z)
        except RuntimeError: # raised by get_xarray_centroid() for empty xarray
            (centr_x, centr_y) = (np.nan, np.nan)
            
        (centr_x, centr_y) = projected_xy_to_longlat((centr_x, centr_y), epsg)
            
    elif centroid=='from_polygon':
        (centr_x, centr_y) = projected_xy_to_longlat(
                                       get_polygon_centroid(polygon), epsg)
    elif isinstance(centroid, tuple):
        centr_x = float(centr_x)
        centr_y = float(centr_y)

    area_per_pixel = np.abs(clips['elevation'].transform[0] * 
                            clips['elevation'].transform[4])/1000000 # in km2
    
    if 'quartz' in XX.keys():
        #print("correcting for quartz distribution")
        # shielding and elevation raster weighted by quartz:
        for key, val in XX.items():
            if key!='quartz':
                XX[key] = val*XX['quartz']
    else:
        # normalize to 1
        XX['quartz'] = XX['elevation']*0+1
    
    # Z denotes elevation raster
    # X denotes any other raster
    df = pd.DataFrame()
    i = 0
    for this_bin in zip(bins[:-1], bins[1:]):
    #for this_bin in zip([bins[1]], [bins[2]]):
        X_bins = {} #dict of rasters for the current bin
        # create mask layer: 1: pixel elevation in this_bin; 0: below this_bin; 2: above this_bin or nan
        Z_bin = xr.apply_ufunc(np.digitize, Z, this_bin)
        # mask each (quartz-weighted) raster to the elevations of this_bin
        for key, val in XX.items():
            X_bins[key] = val.where(Z_bin==1)
            
            
        if np.count_nonzero(~np.isnan(Z_bin))>0:
            temp = pd.DataFrame({'bin': this_bin[0], 'bin1': this_bin[1]}|
                                {key: np.nansum(val) for key, val in X_bins.items()}|
                                {'area_px': np.count_nonzero(~np.isnan(X_bins['elevation'])),
                                 'lat': centr_y, 'long': centr_x}, index = [i])
            df = pd.concat([df, temp], ignore_index=True, sort=False)
            i += 1
            
    for key in XX.keys():
        if key!='quartz':
            df[key] = df[key]/df['quartz']
    df['wt'] = df['quartz']/df['quartz'].sum()
    df['area'] = df['area_px']*area_per_pixel # area in km2
    df = df.loc[df['quartz']>0] # exclude bins with no quarzt contribution
    
    #df.insert(0, 'bins', list(zip(df.bin0, df.bin1)))
    #df.drop(columns=['bin0', 'bin1'], inplace=True)
    df.drop(columns=['bin1', 'area_px', 'quartz'], inplace=True)
        
    summary = {
        'elevLo' : np.nanpercentile(Z, 30),
        'elev50' : np.nanpercentile(Z, 50.),
        'elevHi' : np.nanpercentile(Z, 70),
        'lat' : centr_y,
        'long' : centr_x,
        'areakm2' : sum(df['area'])
        }
    for k, v in clips.items():
        if k not in ('elevation', 'epsg', 'name'):
            summary[k] = np.nanmean(v)
    summary['epsg'] = epsg # avoids that epsg is storead as float
    
    return df, summary


def import_data(df:pd.DataFrame) -> pd.DataFrame:
    """
    Recast pandas DataFrame df for exchange with online calculator. Prints error
    messages; prints warnings if default values are filled in for missing columns.
    Returns an empty DataFrame if problems cannot be resolved.
    
    Parameters
    ----------
    df : pd.DataFrame (see below)
        > df = pd.read_excel('./folder/file.ods') # read from excel spreadsheet
        > df = import_data(df)
                
    Returns
    -------
    out : pd.DataFrame
        columns: {'name', 'lat', 'long', 'elevation', 'press_flag',
                  'thickness', 'density', 'erate', 'shielding', 'year',
                  'nuclide', 'mineral', 'N', 'delN', 'standardization'}

    
    Input DataFrame df must have the following data:
    Sample site information: column names are case-insensitive; several aliases are accepted:
    - 'name'
    - 'lat', 'latitude' :             -90 to +90 decimal degrees
    - 'long', 'longitude', 'lon' :    -180 to +180 decimal degrees
    - 'elevation', 'elev', 'pressure', 'press' :  m or hPa, depending on 'press_flag'
    - 'press_flag':                   'std', 'ant' or 'pre' (standard or antarctica elevation model or pressure in hPa)
    - 'thickness', 'thick' :          sample thickness in cm
    - 'density', 'rho' :              density in g/cm3
    - 'shielding' :                   shielding factor between 0.0 and 1.0
    - 'erate', 'erosion_rate' :       erosion rate in cm/yr
    - 'year' :                        year of sampling betweem 1700 and 2100
    
    Nuclide information: either separate columns for Be and Al data or
    all data in N and nuclide specified in separate column:
    - 'N' :                           nuclide concentration in atoms/g
    - 'delN' (or 'dN', 'N_uncert') :  uncertainty on nuclide concentration in atoms/g
    - 'nuclide' :                     either 'Be-10' or 'Al-26'
    - 'mineral' :                     must be 'quartz'
    - 'standardization', 'standard', 'std' :  see online information
    
    - 'Be' (or 'Be-10', '10Be')
    - 'Be_uncert' (or 'Be_uncertainty' etc)
    - 'Be_standardization' (or 'Be_standard', 'Be_std' etc)
    and/or equivalent for Al
    
    Note that for nuclide information only 'uncertainty', 'standardization' etc
    are case-insensitive. Be, Al and N are case-sensitive.
    
    There are default values if columns 'nuclide', 'mineral' or 'standardization' are missing.
    There are also default values for some of the sample site information; see print_defaults().
    
    """
    # It is recommended to use lowercase letters for all column names except for those referring to nuclide
    # data ('N', 'delN' or 'Be', 'Be_uncert', 'Be_standardization' etc.).
    # All other column names are converted to lowercase, therefore two column names 'name' and 'Name' are
    # considered identical and will cause an error because it is unclear which column to use for sample names.
    
    # missing values raise a ValueError "Non-numeric or missing value in ..."
    # nuclide concentrations <= 0 are silently excluded
    # uncertainties <= 0 raise ValueError "Be uncertainty must be >0."

    #strip leading/trailing spaces from all strings
    df_obj = df.select_dtypes('object')
    df[df_obj.columns] = df_obj.apply(lambda x: x.str.strip())
    
    dfS, dfN = import_data2(df)
    # raises TypeError if input is not a pd.DataFrame
    # prints messages and returns empty dfS, dfN if input is faulty
    
    out = pd.merge(dfS, dfN, on='name', how='right') # duplicates sample info for duplicate analyses
    out = out[ ['name'] + [ col for col in out.columns if col != 'name' ] ] # move 'name' to beginning
    return out


def import_data2(df:pd.DataFrame) -> (pd.DataFrame, pd.DataFrame):
    """
    Recast pandas DataFrame df for exchange with online calculator.
    Returns separate DataFrames for sample and nuclide information.
    See also import_data().
    
    Parameters
    ----------
    df : pd.DataFrame (see below)
        > df = pd.read_excel('./folder/file.ods') # read from excel spreadsheet
        > dfS, dfN = import_data2(df)
                
    Returns
    -------
    dfS : pd.DataFrame
        columns: {'name', 'lat', 'long', 'elevation', 'press_flag',
                  'thickness', 'density', 'erate', 'shielding', 'year'}
        dfS is empty if input data is faulty
    dfN : pd.DataFrame
        columns: {'name', 'nuclide', 'mineral', 'N', 'delN', 'standardization'}
        dfN is empty if input data is faulty
    
    Raises
    ------
    TypeError
        arg is not pd.DataFrames.
        
    """
    
    
    # create pd.DataFrame dfS with sample site information (each sample exactly once)
    # create pd.DataFrame dfN with nuclide information (each sample of dfS at least once)

    # sample site info:
    dfS = pd.DataFrame(columns = ['name', 'lat', 'long', 'elevation',
                                  'press_flag', 'thickness', 'density',
                                  'shielding', 'erate', 'year'])
    dfN = pd.DataFrame(columns = ['name', 'nuclide', 'mineral',
                                  'N', 'delN', 'standardization'])
    
    if isinstance(df, pd.DataFrame):
        temp = df
    else:
        raise TypeError("import_data() argument 'df' must be pandas DataFrame. Try:\n"+
                        "df = pandas.read_excel('file.xlsx')\n"+
                        "df = riversand.import_data(df)")
        
    # fold cases
    keep_cases =  ['Be', 'Be-10', '10Be', 
                   'Al', 'Al-26', '26Al',
                   'N', 'dN', 'delN', 'N_uncert']
    # also uncertainties can be 'Be_uncert' or 'Be_uncert*' (case-insensitive)
    # standardizations can be 'Be_standard' or 'Be_standard*' or 'Be_std' (case-insensitive)
    # note that 'standardisation', 'standardization', 'standard', 'std' are accepted for 'N'
    renames = {}
    for col in temp.columns:
        if col[:2] in ['N_']:
            if col[2:8].casefold()=='uncert':
                renames[col] = col[:2]+'uncert' # uncertainties 'N_uncert...'
        elif col[:3] in ['Be_', 'Al_']:
            if col[3:9].casefold()=='uncert': 
                renames[col] = col[:3]+'uncert' # uncertainties 'Be_uncert...', 'Al_uncert..'
            if col[3:11].casefold()=='standard':
                renames[col] = col[:3]+'standardization' # standardizations 'Be_standard..', 'Al_standard.'
            if col[3:].casefold()=='std':
                renames[col] = col[:3]+'standardization' # standardizations 'Be_std', 'Al_std'
        elif col not in keep_cases:
            renames[col] = col.casefold() # casefold all other column names
    temp = temp.rename(columns=renames)
    if len(temp.columns.unique()) < len(temp.columns):
        print("Non-unique column names.")
        return dfS, dfN
    
            
    # Check that all SAMPLE SITE data are provided; rename columns if necessary
    # name
    if 'name' not in temp.columns:
        print("Missing column 'name'.")
        return dfS, dfN
    else:
        temp['name'] = temp['name'].astype(str)
        if not all([valid_namestring(nm) for nm in temp['name'].to_numpy().tolist()]):
            print("Invalid sample name.")
            return dfS, dfN

    # latitude
    keys = ['latitude', 'lat']
    rename = {k:'lat' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("Missing column 'latitude'.")
        return dfS, dfN
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as 'latitude'.")
        return dfS, dfN
    temp['lat'] = temp['lat'].fillna('nan') # force ValueError for any missing value
    try:
        temp['lat'] = pd.to_numeric(temp['lat'])
    except:
        print("Non-numeric or missing value in 'latitude'.")
        return dfS, dfN
    if ((temp.loc[:,'lat']<(-90)).any() & (temp.loc[:,'lat']>90).any()):
        print("Latitude is out of bounds (-90..+90).")
        return dfS, dfN
    
    # longitude
    keys = ['longitude', 'long', 'lon']
    rename = {k:'long' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("Missing column 'longitude'.")
        return dfS, dfN
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as 'longitude'.")
        return dfS, dfN
    temp['long'] = temp['long'].fillna('nan') # force ValueError for any missing value
    try:
        temp['long'] = pd.to_numeric(temp['long'])
    except:
        print("Non-numeric or missing value in 'longitude'.")
        return dfS, dfN
    if ((temp.loc[:,'long']<(-180)).any() | (temp.loc[:,'long']>180).any()):
        print("Longitude is out of bounds (-180..+180).")
        return dfS, dfN
    
    # elevation
    keys = ['elevation', 'elev', 'pressure', 'press']
    rename = {k:'elevation' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("Missing column 'elevation' or 'pressure'.")
        return dfS, dfN
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as 'elevation' or 'pressure'.")
        return dfS, dfN
    temp['elevation'] = temp['elevation'].fillna('nan') # force ValueError for any missing value
    try:
        temp['elevation'] = pd.to_numeric(temp['elevation'])
    except:
        print("Non-numeric or missing value in 'elevation' or 'pressure'.")
        return dfS, dfN
    if (temp.loc[:,'elevation']<(-500)).any():
        print("Elevation (pressure) is less than -500.")
        return dfS, dfN
    
    #press_flag
    if 'press_flag' in temp.columns:
        if ~temp.loc[:,'press_flag'].isin(['std', 'ant', 'pre']).all():
            print("Invalid 'press_flag'; must be 'std', 'ant' or 'pre'.")
            return dfS, dfN
    else:
        print("No column 'press_flag', assuming default value '{}'."
             .format(Params.default_values['press_flag']))
        temp.loc[:,'press_flag'] = Params.default_values['press_flag']
    
    keys = ['thickness', 'thick']
    rename = {k:'thickness' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("No column 'thickness', assuming default value {} cm."
              .format(Params.default_values['thickness']))
        temp.loc[:,'thickness'] = Params.default_values['thickness']
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as 'thickness'.")
        return dfS, dfN
    temp['thickness'] = temp['thickness'].fillna('nan') # force ValueError for any missing value
    try:
        temp['thickness'] = pd.to_numeric(temp['thickness'])
    except:
        print("Non-numeric or missing value in 'thickness'.")
        return dfS, dfN
    if (temp.loc[:,'thickness']<0).any():
        print("Sample thickness is less than 0.")
        return dfS, dfN
    
    keys = ['density', 'rho']
    rename = {k:'density' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("No column 'density', assuming default value {} g/cm3."
              .format(Params.default_values['density']))
        temp.loc[:,'density'] = Params.default_values['density']
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as 'density'.")
        return dfS, dfN
    temp['density'] = temp['density'].fillna('nan') # force ValueError for any missing value
    try:
        temp['density'] = pd.to_numeric(temp['density'])
    except:
        print("Non-numeric or missing value in 'density'.")
        return dfS, dfN
    if (temp.loc[:,'density']<=0).any():
        print("Density is 0 or less.")
        return dfS, dfN
    
    if 'shielding' in temp.columns:
        temp['shielding'] = temp['shielding'].fillna('nan') # force ValueError for any missing value
        try:
            temp['shielding'] = pd.to_numeric(temp['shielding'])
        except:
            print("Non-numeric or missing value in 'shielding'.")
            return dfS, dfN
    else:
        print("No column 'shielding', assuming default value {:.1f}."
              .format(Params.default_values['shielding']))
        temp.loc[:,'shielding'] = Params.default_values['shielding']
    if ((temp.loc[:,'shielding']<0).any() | (temp.loc[:,'shielding']>1).any()):
        print("Shielding is out of bounds (0..1).")
        return dfS, dfN

    keys = ['erate', 'erosion_rate']
    rename = {k:'erate' for k in keys if k in temp.columns}
    if len(rename)==1:
        temp = temp.rename(columns=rename)
    elif len(rename)==0:
        print("No column 'erate', assuming default value {} cm/yr."
              .format(Params.default_values['erate']))
        temp.loc[:,'erate'] = Params.default_values['erate']
    elif len(rename) > 1:
        print("Multiple columns may be interpreted as erosion rate ('erate').")
        return dfS, dfN
    temp['erate'] = temp['erate'].fillna('nan') # force ValueError for any missing value
    try:
        temp['erate'] = pd.to_numeric(temp['erate'])
    except:
        print("Non-numeric or missing value in 'erate'.")
        return dfS, dfN
    if (temp.loc[:,'erate']<0).any():
        print("Erosion rate is less than 0.")
        return dfS, dfN
    
    if 'year' in temp.columns:
        temp['year'] = temp['year'].fillna('nan') # force ValueError for any missing value
        try:
            temp['year'] = pd.to_numeric(temp['year'], downcast='integer')
        except:
            print("Non-numeric or missing value in 'year'.")
            return dfS, dfN
    else:
        print("No column 'year', assuming default value {}."
              .format(Params.default_values['year']))
        temp.loc[:,'year'] = Params.default_values['year']
    if ((temp.loc[:,'year']<1700).any() | (temp.loc[:,'year']>2100).any()):
        print("Year of collection is out of bounds (1700..2100).")
        return dfS, dfN
    
    dfS = temp[dfS.keys()] # copy sample site columns to dfS
    dfS = dfS[~dfS.duplicated()] # drop rows with identical sample site information
    if dfS['name'].duplicated().any(): # raise error if duplicate names remain
        print("Duplicate samples with non-matching site information (samples: {})."
              .format(', '.join(set(dfS.loc[dfS['name'].duplicated(), 'name']))))
        return dfS, dfN
    
    
    
    
    # if there is 'N' and 'delN' then 'nuclide' defaults to 'Be-10'; 'mineral' is optional
    # or else look for 'Be' and 'Be_uncert' and 'Al' and 'Al_uncert'

    def validate_standardization(temp):
        # raise exception if any standardization is invalid
        col = temp.loc[:,'nuclide']=='Be-10'
        #invalid_std = [v for v in temp.loc[col,'standardization'].values if v not in Params.Be_stds.keys()] #list of invalid standard names
        invalid_std = [i for i,v in temp.loc[col,'standardization'].items() if v not in Params.Be_stds.keys()] #list of indices with invalid standards
        if len(invalid_std)>0:
            raise ValueError("Invalid standardization for Be-10 data (lines: {})."
                            .format(', '.join(str(v) for v in invalid_std)))
        col = temp.loc[:,'nuclide']=='Al-26'
        #invalid_std = [v for v in temp.loc[col,'standardization'].values if v not in Params.Al_stds.keys()]
        invalid_std = [i for i,v in temp.loc[col,'standardization'].items() if v not in Params.Al_stds.keys()]
        if len(invalid_std)>0:
            raise ValueError("Invalid standardization for Al-26 data (lines: {})."
                            .format(', '.join(str(v) for v in invalid_std)))
            
    # Check that NUCLIDE data are provided; rename columns

    # rename nuclide columns to 'Be' and 'Al'; raise exception if 'N' as well as 'Be','Al' are specified
    keys = ['Be', 'Be-10', '10Be']
    rename = {k:'Be' for k in keys if k in temp.columns}
    if len(rename)>1:
        print("Multiple columns may be interpreted as Be concentration.")
        return dfS, dfN
    elif (len(rename)==1) and ('N' in temp.columns):
        print("Multiple columns ('N','Be') may be interpreted as Be concentration.")
        return dfS, dfN
    elif len(rename)==1:
        temp = temp.rename(columns=rename)
        
    keys = ['Be_standardization', 'Be_standardisation', 'Be_standard', 'Be_std']
    rename = {k:'Be_standardization' for k in keys if k in temp.columns}
    if len(rename)>1:
        print("Multiple columns may be interpreted as 'Be_standardization'.")
        return dfS, dfN
    elif len(rename)==1:
        temp = temp.rename(columns=rename)

    keys = ['Al', 'Al-26', '26Al']
    rename = {k:'Al' for k in keys if k in temp.columns}
    if len(rename)>1:
        print("Multiple columns may be interpreted as Al concentration.")
        return dfS, dfN
    elif (len(rename)==1) and ('N' in temp.columns):
        print("Multiple columns ('N','Al') may be interpreted as Al concentration.")
        return dfS, dfN
    elif len(rename)==1:
        temp = temp.rename(columns=rename)
        
    keys = ['Al_standardization', 'Al_standardisation', 'Al_standard', 'Al_std']
    rename = {k:'Al_standardization' for k in keys if k in temp.columns}
    if len(rename) > 1:
        print("Multiple columns may be interpreted as 'Al_standardization'.")
        return dfS, dfN
    elif len(rename)==1:
        temp = temp.rename(columns=rename)


    if 'N' in temp.columns:
        df1 = pd.DataFrame()
        try:
            #temp['N'] = pd.to_numeric(temp['N'])
            df1 = temp.loc[pd.to_numeric(temp.loc[:, 'N'])>0,:]
        except:
            print("Non-numeric value in nuclide concentration 'N'.")
            return dfS, dfN
        
        # N, delN
        keys = ['delN', 'dN', 'N_uncert']
        rename = {k:'delN' for k in keys if k in temp.columns}
        if len(rename)==1:
            df1 = df1.rename(columns=rename)
        elif len(rename)==0:
            print("Missing uncertainty 'delN' for concentration 'N'.")
            return dfS, dfN
        elif len(rename) > 1:
            print("Multiple columns may be interpreted as uncertainty 'delN'.")
            return dfS, dfN
        df1['delN'] = df1['delN'].fillna('nan') # force ValueError for any missing value
        try:
            df1['delN'] = pd.to_numeric(df1['delN'])
        except:
            print("Non-numeric or missing value in nuclide uncertainty 'delN'.")
            return dfS, dfN
        if (df1.loc[:,'delN']<=0).any():
            print("Nuclide uncertainty must be >0.")
            return dfS, dfN
        
        # nuclide
        if 'nuclide' in df1.columns:
            if ~df1.loc[:,'nuclide'].isin(['Be-10','Al-26']).all():
                print("Invalid nuclide; must be 'Be-10' or 'Al-26'.")
                return dfS, dfN
        else:
            print("No column 'nuclide', assuming default '{}'."
                  .format(Params.default_values['nuclide']))
            df1.loc[:,'nuclide'] = Params.default_values['nuclide']

        # mineral
        if 'mineral' in df1.columns:
            if ~df1.loc[:,'mineral'].isin(['quartz']).all():
                print("Invalid mineral; must be 'quartz'.")
                return dfS, dfN
        else:
            #print("No column 'mineral', assuming default mineral '{}'."
            #  .format(Params.default_values['mineral']))
            df1.loc[:,'mineral'] = Params.default_values['mineral']

        # standardization
        keys = ['standardisation', 'standardization', 'standard', 'std']
        rename = {k:'standardization' for k in keys if k in df1.columns}
        if len(rename)==1:
            df1 = df1.rename(columns=rename)
            try:
                validate_standardization(df1)
            except ValueError as e:
                print(e)
                return dfS, dfN
        elif len(rename)==0:
            print("No column 'standardization', assuming default values ({})."
                  .format(', '.join(["{}: {}".format(k,v) for k,v in Params._default_standards.items()])))
            for k,v in Params._default_standards.items():
                col = (df1.loc[:,'mineral']=='quartz') & (df1.loc[:,'nuclide']==k)
                df1.loc[col,'standardization'] = v
        elif len(rename) > 1:
            print("Multiple columns may be interpreted as 'standardization'.")
            return dfS, dfN
        dfN = df1.loc[:, dfN.columns] # df1 has only nuclide-relevant columns and rows
        
        
    elif ('Be' in temp.columns) or ('Al' in temp.columns): # 'mineral' and 'nuclide' are ignored
        dfBe = pd.DataFrame()
        dfAl = pd.DataFrame()
        
        if 'Be' in temp.columns:
            try:
                dfBe = temp.loc[pd.to_numeric(temp.loc[:, 'Be'])>0,:]
            except:
                print("Non-numeric value in Be concentration.")
                return dfS, dfN
            dfBe.loc[:,'mineral'] = 'quartz'
            dfBe.loc[:,'nuclide'] = 'Be-10'
            dfBe.loc[:,'N'] = dfBe.loc[:,'Be']
            try:
                dfBe.loc[:,'delN'] = dfBe.loc[:,'Be_uncert']
            except:
                print("Missing uncertainty 'Be_uncert'.")
                return dfS, dfN
            dfBe['delN'] = dfBe['delN'].fillna('nan') # force ValueError for any missing value
            try:
                dfBe['delN'] = pd.to_numeric(dfBe['delN'])
            except:
                print("Non-numeric or missing value in Be uncertainty.")
                return dfS, dfN
            if (dfBe.loc[:,'delN']<=0).any():
                print("Be uncertainty must be >0.")
                return dfS, dfN
            try:
                dfBe.loc[:,'standardization'] = dfBe.loc[:,'Be_standardization']
            except:
                print("No column 'Be_standardization', assuming default values (Be-10: {})."
                      .format(Params._default_standards['Be-10']))
                dfBe.loc[:,'standardization'] = Params._default_standards['Be-10']
            try:
                validate_standardization(dfBe)
            except ValueError as e:
                print(e)
                return dfS, dfN
            dfBe.index = dfBe.index*2
            
        if 'Al' in temp.columns:
            try:
                dfAl = temp.loc[pd.to_numeric(temp.loc[:, 'Al'])>0,:]
            except:
                print("Non-numeric value in Al concentration.")
                return dfS, dfN
            dfAl.loc[:,'mineral'] = 'quartz'
            dfAl.loc[:,'nuclide'] = 'Al-26'
            dfAl.loc[:,'N'] = dfAl.loc[:,'Al']
            try:
                dfAl.loc[:,'delN'] = dfAl.loc[:,'Al_uncert']
            except:
                print("missing uncertainty 'Al_uncert'.")
                return dfS, dfN
            dfAl['delN'] = dfAl['delN'].fillna('nan') # force ValueError for any missing value
            try:
                dfAl['delN'] = pd.to_numeric(dfAl['delN'])
            except:
                print("Non-numeric or missing value in Al uncertainty.")
                return dfS, dfN
            if (dfAl.loc[:,'delN']<=0).any():
                print("Al uncertainty must be >0.")
                return dfS, dfN
            try:
                dfAl.loc[:,'standardization'] = dfAl.loc[:,'Al_standardization']
            except:
                print("No column 'Al_standardization', assuming default values (Al-26: {})."
                      .format(Params._default_standards['Al-26']))
                dfAl.loc[:,'standardization'] = Params._default_standards['Al-26']
            try:
                validate_standardization(dfAl)
            except ValueError as e:
                print(e)
                return dfS, dfN
            dfAl.index = dfAl.index*2 + 1
            
        
        df1 = pd.concat([dfBe, dfAl], axis=0)#, ignore_index=True)
        dfN = df1.loc[:, dfN.columns] # dfN has only nuclide-relevant columns and rows
        dfN = dfN.sort_index()
    
    else:
        print("Missing columns for nuclide concentrations: either 'N','delN','standardization' "+
              "or 'Be','Be_uncert','Be_std','Al','Al_uncert','Al_std'")
        return dfS, dfN
    
    
    checksum = set(dfN.loc[:,'name']) - set(dfS.loc[:,'name'])
    if len(checksum)>0:
        print("No sample sites info for samples: {} (this error should never occur!!)"
              .format(','.join(sorted(checksum))))
        dfN = dfN[dfN.name.isin(dfS.name)] # drop rows
    checksum = set(dfS.loc[:,'name']) - set(dfN.loc[:,'name'])
    if len(checksum)>0:
        print("No nuclide data for samples: {}."
              .format(', '.join(sorted(checksum))))    
        dfS = dfS[dfS.name.isin(dfN.name)] # drop rows
        
    return dfS, dfN
    
    
    
def get_textblock(df) -> str:
    """
    Get textblock for online calculator. Relies on import_data() for validation.
    Returns an empty string if import_data() encounters problems with the input
    and returns an empty DataFrame.
    
    Parameters
    ----------
    df : pd.DataFrame
        > df = pd.read_excel('./folder/file.ods')   # read from excel spreadsheet
        > df = import_data(df)                      # recast for online calculator / validate
        > textblock = get_textblock(df)             # convert data to string
        or
        > df = pd.read_excel('./folder/file.ods')   # read from excel spreadsheet
        > textblock = get_textblock(df)
                
    Returns
    -------
    textblock : str
        Textblock for online calculator v.3
        
    Raises
    ------
    ValueError
        Raised by validate_topo(), validate_sample(), validate_nuclide()
        e.g. if mandatory item is missing or ill-defined.
        This should never happen because get_textblock() validates the input
        using import_data2().

    """

    
    dfS, dfN = import_data2(df)
        
    dfS = dfS.to_dict(orient='records') # list of dicts, sample site information
    dfN = dfN.to_dict(orient='records') # list of dicts, nuclide information
    
    #import_data2() returns empty df's if data validation failed
    if len(dfS)==0: return ''
    if len(dfN)==0: return ''
    
    # the validation should not be necessary; data is validated by import_data2()
    def lineS(d:dict): # sample and topo data for each row in dfS
        try:
            out1 = validate_topo(d, shielding=True) # 'lat', 'long', 'elevation', 'shielding'
            out3 = validate_sample(d) # 'press_flag', 'density', 'thickness', 'erate', 'year'
        except ValueError as e:
            raise e
        out = {**out1, **out3}
            
        textline = "{} {:.3e} {:.3e} {:.4e} {} {} {} {:.3e} {:.3e} {};".format(
            d['name'], out['lat'], out['long'], out['elevation'], out['press_flag'],
            out['thickness'], out['density'], out['shielding'], out['erate'], out['year'])
        return textline
    
    # the validation should not be necessary; data is validated by import_data2()
    def lineN(d:dict): # nuclide data for each row in dfN
        try:
            out = validate_nuclide(d) # 'name', 'N', 'delN', 'nuclide', 'mineral', 'standardization'
        except ValueError as e:
            raise e
            
        textline = "{} {} {} {} {} {};".format(
            out['name'], out['nuclide'], out['mineral'], out['N'], out['delN'], out['standardization'])
        return textline
    
    textblockS = [lineS(d) for d in dfS]
    textblockN = [lineN(d) for d in dfN]
    textblockS = ' '.join(textblockS)
    textblockN = ' '.join(textblockN)
    return textblockS + ' ' + textblockN



__version__ = "1.3.1"

# function get_bins()
# set dtype=int to fix DeprecationWarning __array__

# updated quartz weighing function

# fixed rounding error for low erosion rates:
#    get_textline(), get_textblock()
# remove leading and trailing spaces in data from spreadsheet file
#    .add_samples_from_dataframe(), .add_samples_from_file(), import_data()
# plotting functions returning fig, ax

# print_defaults(), print_standardizations()
# restandardize()

# Point-based erosion rates and exposure ages
# import_data(), get_textblock()
# get_ages_from_server(), get_erates_from_server()


